#ifdef LIFE

#include "stock.h"

int rows_tot;
int cols_tot;

char **cell;

#define ALIVE 1
#define DEAD  0

static void calculate_dividends ();
static void deliver_dividends ();

void initiate_game_of_life () {
   register int i,j;
   
   rows_tot = max_companies + 2;
   cols_tot = max_groups + 2;
   
   /* nifty trick for dynamically assigning contiguous chunk of memory
    * for a 2d array 
    */
   
   cell = (char **) malloc (rows_tot * sizeof (char *));
   cell[0] = (char *) malloc (rows_tot * cols_tot * sizeof(char));
   for (i=1; i<rows_tot; ++i)
     cell[i] = cell[0] + i * cols_tot;
   
   for (i=0; i<rows_tot; ++i)
     for (j=0; j<cols_tot; ++j) 
       cell[i][j] = rand () % 2;
   
   
}

void iteration () {
   register int i, j;
   int rows = rows_tot - 1;
   int cols = cols_tot - 1;
   
   for (i=1; i<rows; ++i) {
      for (j=1; j<cols; ++j) {
	 register int t = 0;
	 
	 if (cell[i+1][j+1] == ALIVE) ++t;
	 if (cell[i+1][j] == ALIVE) ++t;
	 if (cell[i+1][j-1] == ALIVE) ++t;
	 
	 if (cell[i][j+1] == ALIVE) ++t;
	 if (cell[i][j-1] == ALIVE) ++t;
	 
	 if (cell[i-1][j+1] == ALIVE) ++t;
	 if (cell[i-1][j] == ALIVE) ++t;
	 if (cell[i-1][j-1] == ALIVE) ++t;
	 
	 if ((t < 2) || (t > 4)) cell[i][j] = DEAD;
	 else cell[i][j] = ALIVE;
      }
   }

   for (j=1; j<cols; ++j) {
      register int t=0;
      
      if (cell[1][j-1] == ALIVE) ++t;
      if (cell[1][j] == ALIVE) ++t;
      if (cell[1][j+1] == ALIVE) ++t;
      
      if (cell[0][j-1] == ALIVE) ++t;
      if (cell[0][j+1] == ALIVE) ++t;
      
      if (cell[rows][j-1] == ALIVE) ++t;
      if (cell[rows][j] == ALIVE) ++t;
      if (cell[rows][j+1] == ALIVE) ++t;
      
      if ((t < 2) || (t > 4)) cell[i][j] = DEAD;
      else cell[i][j] = ALIVE;
      
      t = 0;
      
      if (cell[rows-1][j-1] == ALIVE) ++t;
      if (cell[rows-1][j] == ALIVE) ++t;
      if (cell[rows-1][j+1] == ALIVE) ++t;
      
      if (cell[rows][j-1] == ALIVE) ++t;
      if (cell[rows][j+1] == ALIVE) ++t;
      
      if (cell[0][j-1] == ALIVE) ++t;
      if (cell[0][j] == ALIVE) ++t;
      if (cell[0][j+1] == ALIVE) ++t;
      
      if ((t < 2) || (t > 4)) cell[i][j] = DEAD;
      else cell[i][j] = ALIVE;
      
   }

   for (i=1; i<rows; ++i) {
      register int t = 0;
      
      if (cell[i+1][cols] == ALIVE) ++t;
      if (cell[i][cols] == ALIVE) ++t;
      if (cell[i-1][cols] == ALIVE) ++t;
      
      if (cell[i+1][0] == ALIVE) ++t;
      if (cell[i-1][0] == ALIVE) ++t;
      
      if (cell[i+1][1] == ALIVE) ++t;
      if (cell[i][1] == ALIVE) ++t;
      if (cell[i-1][1] == ALIVE) ++t;
      
      if ((t < 2) || (t > 4)) cell[i][j] = DEAD;
      else cell[i][j] = ALIVE;

      t = 0;
      
      if (cell[i+1][cols-1] == ALIVE) ++t;
      if (cell[i][cols-1] == ALIVE) ++t;
      if (cell[i-1][cols-1] == ALIVE) ++t;
      
      if (cell[i+1][cols] == ALIVE) ++t;
      if (cell[i-1][cols] == ALIVE) ++t;
      
      if (cell[i+1][0] == ALIVE) ++t;
      if (cell[i][0] == ALIVE) ++t;
      if (cell[i-1][0] == ALIVE) ++t;
      
      if ((t < 2) || (t > 4)) cell[i][j] = DEAD;
      else cell[i][j] = ALIVE;

   }
   
   if (rand () % 2 == 0)
     cell[0][0] = ALIVE;
   else cell[0][0] = DEAD;

   if (rand () % 2 == 0)
     cell[0][cols] = DEAD;
   else cell[0][cols] = ALIVE;
   
   if (rand() % 2 == 0)
     cell[rows][0] = DEAD;
   else cell[rows][0] = ALIVE;
   
   if (rand () % 2 == 0)
     cell[rows][cols] = DEAD;
   else cell[rows][cols] = ALIVE;

}


void *life () {
   initiate_game_of_life ();
   
   while (1) {
      int time_to_dividends = 60;
            
      iteration ();
      if (--time_to_dividends == 0) {
	 register plid_t plid;
	 
	 time_to_dividends = 60;
	 
	 calculate_dividends ();
	 
	 deliver_dividends ();
      }
      
      sleep (admin_interval);
   }
}

static void calculate_dividends () {
   register cid_t cid;
   
   for (cid=0; cid <= max_companies; ++cid) {
      int i = cid + 1, g = company[cid].gid + 1, dividend = 0;

      if (cell[i][g] == DEAD) {
	 dividend = 0;
	 continue;
      }

      if (cell[i+1][g+1] == ALIVE) ++dividend;
      if (cell[i+1][g] == ALIVE) ++dividend;
      if (cell[i+1][g-1] == ALIVE) ++dividend;
      
      if (cell[i][g+1] == ALIVE) ++dividend;
      if (cell[i][g-1] == ALIVE) ++dividend;
      
      if (cell[i-1][g+1] == ALIVE) ++dividend;
      if (cell[i-1][g] == ALIVE) ++dividend;
      if (cell[i-1][g-1] == ALIVE) ++dividend;
      
      company[cid].dividend = dividend * 2;
   }
}

void deliver_dividends () {
   plid_t plid;
   player_t *p;
   stock_player_t **s;
   cash_t dividend_tot;
   register int i;
   
   /* No dividend for plid 0, as that is public */
   
#ifdef LOG
   fprintf (logfile, "Delivering Dividends\n");
#endif
   
#ifdef DEBUG
   printf ("Dividend time\n");
#endif
   
   for (plid=1; plid<=max_players; ++plid) {
      p = &player[plid];
      s = p->holding;
      
      for (i=0; i<hashsize; ++i) {
	 
	 while (s[i] != NULL) {
	    dividend_tot += company[s[i]->cid].dividend * s[i]->qty / 100;
	    s[i] = s[i]->next;
	 }
      }
      
      pthread_mutex_lock (&p->player_mut);
      p->cash += dividend_tot;
      pthread_mutex_unlock (&p->player_mut);
   
      addto_result_list (plid, DIVIDEND, 0, 0, dividend_tot);

   }
}

#endif /* LIFE */
